/*
 * $Id: getopt.c,v 1.8 1995/05/11 20:46:12 vons Exp $
 *
 * POSIX.2 compatible getopt()	- get option letter from argument vector
 *
 * SYNOPSIS
 *     int getopt(int argc, char * const *argv, const char *optstring);
 *
 *     extern char *optarg;
 *     extern int optind, opterr, optopt;
 *
 * $Log: getopt.c,v $
 * Revision 1.8  1995/05/11  20:46:12  vons
 * renamed PROTO to HAVE_PROTOTYPES
 *
 * Revision 1.7  1994/07/30  10:34:59  vons
 * Fixed compiler warning (redundant decl's)
 *
 * Revision 1.6  1994/07/27  17:54:53  vons
 * bugfix: retval for missing arg is ':' or '?' depending on opstring[0]
 *
 * Revision 1.5  1994/06/14  22:12:45  vons
 * Added getopt.h
 *
 * Revision 1.4  1994/04/13  20:12:35  vons
 * Added RCS Id
 *
 * Revision 1.3  1994/04/12  19:59:57  vons
 * fixed lint warning
 *
 * Revision 1.2  1994/04/12  19:25:30  vons
 * Now posix.2 compatible, added optopt support.
 *
 * Revision 1.1  1992/03/02  19:51:56  vons
 * Initial revision
 *
 */

#include <stdio.h>
#include <string.h>

#define __GETOPT_PRIVATE__

#include "getopt.h"

int       optind = 1;             /* start with argv[1]		 */
int       opterr = 1;             /* default error msgs on	 */
char     *optarg = NULL;
int       optopt = '\0';          /* option char found		 */

#ifndef HAVE_PROTOTYPES
#ifdef __STDC__
#define HAVE_PROTOTYPES	1
#endif
#endif

#ifdef HAVE_PROTOTYPES
#define PROTOTYPE(func, args)	func args
#else
#define PROTOTYPE(func, args)	func ()
#define const
#endif

PROTOTYPE(int getopt, (int, char * const *, const char *));
PROTOTYPE(static void _opt_err, (const char *, const char *, int));

static char nullstr[] = "";

static void
_opt_err(argv0, emsg, opt_char)
const char *argv0;
const char *emsg;
int       opt_char;
{
    if (opterr != 0)
        (void) fprintf(stderr, "%s: %s -- '%c'\n", argv0, emsg, opt_char);
}

int
getopt(argc, argv, optstring)
int       argc;
char     *const * argv;
const char *optstring;
{
    static char *opt = nullstr;
    char     *oi;

    if (*opt == '\0')
    {

        /*
         * move to next element in argv, as specified by optind it should
         * exist, and start with a '-'.
         */
        if ((optind >= argc) || (argv[optind][0] != '-'))
        {
            /* end of option list reached */
            return -1;
        }

        /*
         * argument starting with '--' is special, it indicates the end of
         * the option list. Note that this argument is skipped.
         */
        opt = &argv[optind][1];
        if (*opt == '-')
        {
            ++optind;
            opt = nullstr;
            return -1;
        }

        /*
         * If this argv element is only "-", regard it as the '-' option
         * character (for compatibility reasons).
         */
        if (*opt == '\0')
            --opt;
    }

    /*
     * now opt points to option character to check. It is invalid if it is
     * a ':', or if it does not occur in the optstring.
     */

    optopt = *opt++;
    oi = strchr(optstring, (int) optopt);
    if ((optopt == ':') || (oi == NULL))
    {

        /*
         * Caller didn't specify '-' as valid option character. Regard it
         * as the end-of-options (for compatibility reasons). Note that we
         * also arrive here if current argv == "-".
         */
        if (optopt == '-')
            return -1;

        /*
         * Invalid option character found, generate an error message.
         */
        _opt_err(argv[0], "illegal option", optopt);

        /*
         * Move to next option (will be checked at next call), and return
         * the ? character.
         */
        if (*opt == '\0')         /* end of this argv */
            ++optind;

        return '?';
    }

    /* option character is valid, check the option argument */
    if (oi[1] != ':')
    {
        /* move to next arg, used in next call */
        if (*opt == '\0')
            ++optind;

        /* no argument for this option */
        optarg = nullstr;

        return optopt;
    }

    /*
     * There must be an argument. It is either attached to the option, or
     * separated by whitespace.
     */
    if (*opt != '\0')
    {
        /* argument attached, adapt optind and opt, return addr of arg */
        optarg = opt;
        opt = nullstr;
        ++optind;

        return optopt;
    }

    /* argument must be in next argv element, if there is one */
    if (++optind < argc)
    {
        optarg = argv[optind];
        opt = nullstr;
        ++optind;                 /* skip the arg */

        return optopt;
    }

    /* no argument for option, generate error message */
    _opt_err(argv[0], "option requires an argument", optopt);

    /* there is nothing to continue with in next call */
    opt = nullstr;

    if (optstring[0] == ':')
        return ':';
    else
        return '?';
}
