/*
 * $Id: vd.c,v 1.12 1995/08/06 09:08:30 vons Exp $
 */

#include "config.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

/*--------------------------------------------------------------------
 * Atari ST
 *
 * Some compilers/environments need a special call to get stdin or
 * stdout in binary mode. The BINARY_MODE(FILE *fp) macro can be
 * defined if needed.
 */
#ifdef LATTICE
#ifdef ATARI
#define BINARY_MODE(f)	 fmode(f, 1)
#endif
#endif

/*--------------------------------------------------------------------
 * Generic stuff
 */

#ifdef __STDC__
#define HAVE_PROTOS
#endif

#ifdef HAVE_PROTOS
#define PROTOTYPE(func,args)	func args
#else
#define PROTOTYPE(func,args)	func ()
#endif

typedef unsigned long Ulong;

/*
 * Prototypes of internal and external functions.
 */
#ifndef HAVE_POSIX_GETOPT
#include "getopt.h"
#endif

PROTOTYPE(static void usage, (void));
PROTOTYPE(static int bin_to_dump, (void));
PROTOTYPE(static int dump_to_bin, (void));
PROTOTYPE(int main, (int argc, char **argv));

PROTOTYPE(static int hole_fill, (Ulong naddr, Ulong addr, Ulong line_no));
PROTOTYPE(static int hole_seek, (Ulong naddr, Ulong addr, Ulong line_no));
PROTOTYPE(static int hole_fail, (Ulong naddr, Ulong addr, Ulong line_no));
PROTOTYPE(static int hole_ignore, (Ulong naddr, Ulong addr, Ulong line_no));

/* desired function: */
enum Action
{
    Dump, Restore, Restore_no_check, Restore_fill_holes, Restore_seek_holes
};
static enum Action action = Dump;

static const char *prgname;

int
main(argc, argv)
int       argc;
char     *argv[];
{
    int       opt;
    int       retval;
    char     *infile = NULL;
    char     *outfile = NULL;

    /*
     * Use default for those without a proper argv[0]. Else check for both
     * '/' (Unix) and '\' (DOS/TOS).
     */

    if (NULL == argv[0] || 0 == argv[0][0])
        prgname = "vd";
    else if (NULL == (prgname = strrchr(argv[0], '/'))
             && NULL == (prgname = strrchr(argv[0], '\\')))
        prgname = argv[0];
    else
        prgname++;

    opterr = 0;
    while ((opt = getopt(argc, argv, "fFno:rRV")) != EOF)
    {
        switch (opt)
        {
          case 'f':
            action = Restore_fill_holes;
            break;
          case 'F':
            action = Restore_seek_holes;
            break;
          case 'n':              /* Native locale */
#ifdef LC_CTYPE
            if (setlocale(LC_CTYPE, "") == NULL)
            {
                (void) fprintf(stderr,
                               "%s: WARNING, couldn't set locale\n",
                               prgname);
            }
#endif
            break;
	  case 'o':
    	    outfile = optarg;
	    break;
          case 'r':
            action = Restore;
            break;
          case 'R':
            action = Restore_no_check;
            break;
          case 'V':
            (void) fprintf(stderr, "%s: version 1.8\n", prgname);
            return 0;
            /* NOTREACHED */
          case '?':
            (void) fprintf(stderr, "%s: unknown option '%c'\n",
                           prgname, optopt);
            usage();
            return 1;
            /* NOTREACHED */
          default:
            (void) abort();       /* should not occur */
        }
    }


    /* If there is an argument left, it indicates the input file */
    infile = (optind < argc) ? argv[optind++] : NULL;

    /* More arguments indicate an error */
    if (optind != argc)
    {
        (void) fprintf(stderr, "%s: too many arguments\n", prgname);
        usage();
        return 1;
    }

    if (infile != NULL)
    {
        if (freopen(infile, (action == Dump ? "rb" : "r"), stdin) == NULL)
        {
            (void) fprintf(stderr,
                           "%s: couldn't open file %s for reading\n",
                           prgname, infile);
            return 2;
        }
    }
#ifdef BINARY_MODE
    else
    {
        if (Dump == action)
            BINARY_MODE(stdin);
    }
#endif

    /* If an output file has been specified, open it. Else use stdout */
    if (NULL != outfile)
    {
        if (strcmp(infile, outfile) == 0)
        {
            (void) fprintf(stderr,
                           "%s: input and output file are identical\n",
                           prgname);
            return 1;

        }

        if (freopen(outfile, (action == Dump ? "w" : "wb"), stdout) == NULL)
        {
            (void) fprintf(stderr, "%s: couldn't open file %s for writing\n",
                           prgname, outfile);
            return 2;
        }
    }
#ifdef BINARY_MODE
    else
    {
        if (Dump != action)
            BINARY_MODE(stdout);
    }
#endif

    if (Dump == action)
        retval = bin_to_dump();
    else
        retval = dump_to_bin();

    /* Don't care if close() fails on input file */
    (void) fclose(stdin);

    /* Make sure file is written ok */
    if (fclose(stdout) != 0)
    {
        (void) fprintf(stderr,
                     "%s: close failed on output file (write error ?)\n",
                       prgname);
        return 2;
    }

    /*-
     * retval is correctly set by conversion functions.
     *     0: ok, 1: user error, 2: others
     */
    return retval;
}

static void
usage()
{
    (void) fprintf(stderr,
            "Usage: %s [-fFnrRV] [-o output-file] [input-file]\n", prgname);
}

/*
 * Dump output format is as follows:
 *
 *   index wrt ':'      meaning
 *  ---------------+------------------------------
 *  (-10)-(-1)     |	hex address/offset, without leading zero's
 *    0-1	   |	': '	(separator + first space)
 *    2-4	   |	'XX '	(1st hex digit + space)
 *   ...	   .	...
 *   23-25	   |	'XX '	(8th hex digit + space)
 *    26	   |	' '	(additional space separator)
 *   27-29	   |	'XX '	(9th hex digit + space)
 *   ...	   .	...
 *   48-50	   |	'XX '	(16th hex digit + space)
 *   51-52	   |	'| '	(vertical bar + space)
 *   53-68	   |	16 characters
 *   69-70	   |	'\n\0'	(trailing newline + string terminator)
 *
 * All the spaces, separators and all that are generated only once.
 * Wether or not the size of the address grows is not important, since it
 * grows leftward from the ':', it does not force a shift of the whole
 * line when it gets wider.
 */
static int
bin_to_dump()
{
    static char linebuf[81];
    static char to_hex[] = "0123456789ABCDEF";
    static char to_print[256];

    register char *phex, *pascii;
    register int byte, byte_cnt;
    char     *psep, *plinestart;

    Ulong     address = 0;
    int       i;

    /* Build the to_print table. isprint() is locale dependent */
    for (i = 0; i <= 255; i++)
        to_print[i] = (char) (isprint(i) ? i : '.');

    /* generate the constant parts of the string only once */
    (void) memset(linebuf, ' ', 80);
    psep = linebuf + 10;
    psep[0] = ':';
    psep[51] = '|';
    psep[69] = '\n';
    psep[70] = '\0';

    for (;;)
    {

        /*
         * Prefix line with address, which should occupy at least six
         * places. Since address only grows, no need to write spaces in
         * front of it, is done once (see above).
         *
         * Always generate at least one digit, so loop once in all cases
         */
        plinestart = psep - 1;
        {
            register Ulong addr = address;

            do
            {
                *plinestart = to_hex[(int) (addr & (Ulong) 0x0f)];
                addr >>= 4;
                plinestart--;
            }
            while (addr);
        }

        /* move forward to 1st digit */
        plinestart++;

        /* at least 6 digits */
        if (plinestart > psep - 6)
            plinestart = psep - 6;

        phex = psep + 2;
        pascii = psep + 53;

        byte_cnt = 8;
        while (byte_cnt && (byte = getchar()) != EOF)
        {
            *phex = to_hex[byte >> 4];
            phex++;
            *phex = to_hex[byte & 0x0F];
            phex += 2;
            *(pascii++) = to_print[byte];

            byte_cnt--;
        }

        /* Go on if we didn't break out on EOF */
        if (0 == byte_cnt)
        {
            /* separation between first and second 8 bytes */
            phex++;

            byte_cnt = 8;
            while (byte_cnt && (byte = getchar()) != EOF)
            {
                *phex = to_hex[byte >> 4];
                phex++;
                *phex = to_hex[byte & 0x0F];
                phex += 2;
                *(pascii++) = to_print[byte];

                byte_cnt--;
            }
        }
        else
        {

            /*
             * Encountered EOF in first section. If no bytes where
             * processed, break out. Otherwise, fill remainder with spaces
             */
            if (8 == byte_cnt)
                break;

            for (; byte_cnt; byte_cnt--)
            {
                *phex = ' ';
                phex++;
                *phex = ' ';
                phex += 2;

                *(pascii++) = ' ';
            }

            /* separation between first and second 8 bytes */
            phex++;

            /* second part will be 'spaced' below */
            byte_cnt = 8;
        }

        /* Complete second 8 bytes with spaces if needed */
        for (; byte_cnt; byte_cnt--)
        {
            *phex = ' ';
            phex++;
            *phex = ' ';
            phex += 2;

            *(pascii++) = ' ';
        }

        address += (Ulong) 16;

        if (fputs(plinestart, stdout) == EOF)
        {
            (void) fprintf(stderr, "%s: write error (disk full?)\n", prgname);
            return 2;
        }


        /* Get out if we've just written the last (possibly partial) line */
        if (EOF == byte)
            break;
    }

    /* Everything is ok if we've hit eof */
    if (feof(stdin))
        return 0;

    /* Not eof means error */
    (void) fprintf(stderr, "%s: read error\n", prgname);
    return 2;
}

#define LINELEN 256

static int
dump_to_bin()
{
    Ulong     line_no;
    Ulong     address, new_address;
    int       i;

    register unsigned char byte;

    static int to_bin[256];
    static unsigned char line[LINELEN];
    register unsigned char *lp;

    PROTOTYPE(int (*handle_hole), (Ulong naddr, Ulong addr, Ulong lineno));

    /*
     * Create a table to convert hex digits to numbers. By Giving all the
     * non-xdigit entries a value < 0, the array can also be used in place
     * of the isxdigit() function/macro.
     */
    for (i = 0; i <= 255; i++)
        to_bin[i] = -1;

    for (i = '0'; i <= '9'; i++)
        to_bin[i] = i - '0';

    for (i = 'a'; i <= 'f'; i++)
        to_bin[i] = to_bin[toupper(i)] = 10 + i - 'a';

#define Isxdigit(i)	(to_bin[i] >= 0)

    /* how to handle a hole in the address sequence: */
    switch (action)
    {
      case Restore:
        handle_hole = hole_fail;
        break;
      case Restore_no_check:
        handle_hole = hole_ignore;
        break;
      case Restore_seek_holes:
        handle_hole = hole_seek;
        break;
      case Restore_fill_holes:
        handle_hole = hole_fill;
        break;
      default:
        /* should never come here */
        (void) fprintf(stderr,
                 "%s: BUG ALERT, took invalid case in switch, line %i\n",
                       prgname, __LINE__);
        return 2;
    }

    line_no = 0;
    address = 0;

    while (fgets((char *) line, LINELEN, stdin) != NULL)
    {
        line_no++;

        /*
         * Input line is too long if it has max length and no trailing \n.
         * But the last line of the file may be of max length and lack the
         * \n, so check for EOF as well.
         */
        if (strlen((char *) line) == LINELEN - 1 &&
            line[LINELEN - 2] != '\n' &&
            !feof(stdin))
        {
            (void) fprintf(stderr,
                           "%s: line %li too long \n", prgname, line_no);
            return 1;
        }

        lp = line;

        /* skip leading white space */
        while (*lp == ' ' || *lp == '\t')
            lp++;

        if (!Isxdigit(*lp))
        {
            (void) fprintf(stderr, "%s: missing address on line %li\n",
                           prgname, line_no);
            return 1;
        }

        new_address = (Ulong) 0;

        do
            new_address = (new_address << 4) | to_bin[*(lp++)];
        while (Isxdigit(*lp));

        /* If there is a mismatch, handle it properly */
        if (new_address != address)
        {
            int       retval;

            if ((retval = handle_hole(new_address, address, line_no)) != 0)
                return retval;

            address = new_address;
        }

        /* skip optional white space before ':' */
        while (' ' == *lp || '\t' == *lp)
            lp++;


        /* This must be the colon */
        if (':' == *lp)
            lp++;
        else
        {
            (void) fprintf(stderr, "%s: colon expected on line %li\n",
                           prgname, line_no);
            return 1;
        }

        /* skip any white space following the ':' */
        while (' ' == *lp || '\t' == *lp)
            lp++;

        /*
         * Process HH pairs until either the separator is found, or
         * eol/eof is encountered . Note: last line may be terminated by a
         * null, if it lacks the trailing newline!
         */
        while ('|' != *lp && '\n' != *lp && 0 != *lp)
        {

            if (!Isxdigit(*lp))
            {
                (void) fprintf(stderr, "%s: found junk on line %li\n",
                               prgname, line_no);
                return 1;
            }

            byte = (unsigned char) (to_bin[*lp] << 4);
            lp++;

            if (!Isxdigit(*lp))
            {
                (void) fprintf(stderr,
                            "%s: missing second hex digit on line %li\n",
                               prgname, line_no);
                return 1;
            }

            byte = (unsigned char) (byte | to_bin[*lp]);

            if (putchar((int) byte) == EOF)
            {
                (void) fprintf(stderr, "%s: write error (disk full?)\n",
                               prgname);
                return 2;
            }

            address++;

            /*
             * Skip any intermediate whitespace, there must be at least
             * one space/tab.
             */
            lp++;
            if (' ' != *lp && '\t' != *lp)
                break;

            do
                lp++;
            while (' ' == *lp || '\t' == *lp);
        }

        /* It's ok if we've hit the end of the line, or the '|' */

        if ('|' == *lp || '\n' == *lp || 0 == *lp)
            continue;

        if (Isxdigit(*lp))
            (void) fprintf(stderr,
                           "%s: missing space on line %li\n",
                           prgname, line_no);
        else
            (void) fprintf(stderr,
                           "%s: illegal character '%c' on line %li\n",
                           prgname, *lp, line_no);

        return 1;
    }

    /* Everything is ok if we've hit eof */
    if (feof(stdin))
        return 0;

    /* Not eof means error */
    (void) fprintf(stderr, "%s: read error\n", prgname);
    return 2;
}

static int
hole_fill(naddr, addr, line_no)
Ulong     naddr, addr, line_no;
{
    if (naddr < addr)
    {
        (void) fprintf(stderr, "%s: address too small on line %li\n",
                       prgname, line_no);
        return 1;
    }
    for (; addr < naddr; addr++)
    {
        if (putchar('\0') == EOF)
        {
            (void) fprintf(stderr, "%s: write error (disk full?)\n",
                           prgname);
            return 2;
        }
    }

    return 0;
}
static int
hole_seek(naddr, addr, line_no)
Ulong     naddr, addr, line_no;
{
    if (naddr < addr)
    {
        (void) fprintf(stderr, "%s: address too small on line %li\n",
                       prgname, line_no);
        return 1;
    }
    if (fseek(stdout, (long) (naddr - addr), SEEK_CUR) != 0)
    {
        (void) fprintf(stderr, "%s: could not fill hole on line %li\n",
                       prgname, line_no);
        return 2;
    }

    return 0;
}

/*ARGSUSED*/
static int
hole_ignore(naddr, addr, line_no)
Ulong     naddr, addr, line_no;
{
    return 0;
}

/*ARGSUSED*/
static int
hole_fail(naddr, addr, line_no)
Ulong     naddr, addr, line_no;
{
    (void) fprintf(stderr, "%s: address mismatch on line %li\n",
                   prgname, line_no);
    return 1;
}
